""" 
"""
import torch
from .model import RNABertModel
from .utils import load_hub_workaround, token_to_index

MODEL_URL = "https://zenodo.org/record/8263889/files/ncRNABert.pt"


def load_ncRNABert():
    model_data = load_hub_workaround(MODEL_URL)
    # model_data = torch.load("/home/wanglei/data/RNALM/src/ncRNABert.pt")
    hyper_parameter = model_data["hyper_parameters"]
    model = RNABertModel(num_token=hyper_parameter['num_tokens'],
                         num_layers=hyper_parameter['num_layer'],
                         embed_dim=hyper_parameter['embed_dim'],
                         attention_heads=hyper_parameter['attention_heads'],
                         padding_idx=token_to_index["PADDING_MASK"])

    model.load_state_dict(model_data['state_dict'])

    return model


if __name__ == "__main__":
    model = load_ncRNABert()



""" 
"""

import torch.nn as nn
import torch
from einops import rearrange

from .modules import RobertaLMHead, TransformerLayer
from .utils import length_to_mask


class RNABertModel(nn.Module):
    def __init__(
        self,
        num_token=14,
        num_layers: int = 24,
        embed_dim: int = 1024,
        attention_heads: int = 16,
        dropout: float = 0.,
        padding_idx:int = 13
    ):
        super().__init__()
        self.num_token = num_token
        self.padding_idx = padding_idx
        self.num_layers = num_layers
        self.embed_dim = embed_dim
        self.attention_heads = attention_heads
        self.dropout = dropout

        self._init_submodules()

    def _init_submodules(self):
        self.embed_tokens = nn.Embedding(
            self.num_token,
            self.embed_dim,
            padding_idx=self.padding_idx,
        )

        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    self.embed_dim,
                    4 * self.embed_dim,
                    self.attention_heads,
                    self.dropout,
                    self.dropout
                )
                for _ in range(self.num_layers)
            ]
        )

        self.emb_layer_norm_after = nn.LayerNorm(self.embed_dim)

        self.lm_head = RobertaLMHead(
            embed_dim=self.embed_dim,
            output_dim=self.num_token,
            weight=self.embed_tokens.weight
        )

    def forward(self, tokens, lengths=None, repr_layers=[], need_head_weights=False):

        assert tokens.ndim == 2
        x = self.embed_tokens(tokens)

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            attn_weights = []

        attn_mask = length_to_mask(lengths)
        attn_mask = rearrange(attn_mask, 'b l -> b () () l') * rearrange(attn_mask, 'b l -> b () l ()')

        for layer_idx, layer in enumerate(self.layers):
            x, attn = layer(x, attn_mask)
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x
            if need_head_weights:
                # (H, B, T, T) => (B, H, T, T)
                attn_weights.append(attn)
                
        x = self.emb_layer_norm_after(x)

        # last hidden representation should have layer norm applied
        if (layer_idx + 1) in repr_layers:
            hidden_representations[layer_idx + 1] = x

        x = self.lm_head(x)

        result = {"logits": x, "representations": hidden_representations}
        
        if need_head_weights:
            result["attentions"] = attn_weights

        return result


""" 
"""
from typing import Sequence, Tuple, List, Union
from torch.nn.utils.rnn import pad_sequence
import torch
import pathlib
import urllib
import re


nucleotides = "GAUC"
rna_3 = [i + j + k for i in nucleotides for j in nucleotides for k in nucleotides]
rna_2 = [i + j for i in nucleotides for j in nucleotides]
rna_1 = [i for i in nucleotides]

special_rna_tokens = ['<UNK>', 'TOKEN_MASK', 'PADDING_MASK']

rna_tokens = rna_3 + rna_2 + rna_1 + special_rna_tokens


token_to_index = {}
for i, j in enumerate(rna_tokens):
    token_to_index[j] = i

def length_to_mask(length, max_len=None, dtype=None):
    """length: B.
    return B x max_len.
    If max_len is None, then max of length will be used.
    """
    assert len(length.shape) == 1, 'Length shape should be 1 dimensional.'
    max_len = max_len or length.max().item()
    mask = torch.arange(max_len, device=length.device,
                        dtype=length.dtype).expand(len(length), max_len) < length.unsqueeze(1)
    if dtype is not None:
        mask = torch.as_tensor(mask, dtype=dtype, device=length.device)
    return mask

def split_rna_seq(sequence):
    sequence_length = len(sequence)
    token_split_list = [sequence[i:i+3] for i in range(sequence_length-2)]
    token_split_list += [sequence[-2:], sequence[-1]]
    return token_split_list

def BatchConverter(raw_batch: Sequence[Tuple[str, str]]):
    ids = [item[0] for item in raw_batch]
    seqs = [item[1] for item in raw_batch]
    lengths = torch.tensor([len(item[1]) for item in raw_batch])
    batch_token = []
    for seq in seqs:
        seq = re.sub(r"[T]", "U", seq.upper())
        batch_token.append(torch.tensor([token_to_index.get(i, token_to_index["<UNK>"]) for i in split_rna_seq(seq)]))
    batch_token = pad_sequence(batch_token, batch_first=True, padding_value=token_to_index['PADDING_MASK'])
    return ids, batch_token, lengths


def load_hub_workaround(url):
    try:
        data = torch.hub.load_state_dict_from_url(url, progress=False, map_location="cpu")
    except RuntimeError:
        # Pytorch version issue - see https://github.com/pytorch/pytorch/issues/43106
        fn = pathlib.Path(url).name
        data = torch.load(
            f"{torch.hub.get_dir()}/checkpoints/{fn}",
            map_location="cpu",
        )
    except urllib.error.HTTPError as e:
        raise Exception(f"Could not load {url}, check your network!")
    return data


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='ncRNABert',
      version='0.1.2',
      description='ncRNA language model',
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords='ncRNA language model',
      classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
      url='https://github.com/wangleiofficial/ncRNABert',
      author='Lei Wang',
      author_email='wanglei@isyslab.org',
      license='MIT',
      packages=['ncRNABert'],
      install_requires=[
        'einops',
        'torch',
        'numpy',
        'rotary_embedding_torch'
    ],
      include_package_data=True,
      zip_safe=False)


import torch
import torch.nn as nn
from torch import Tensor
from typing import Optional, Tuple
import torch.nn.functional as F
from einops import rearrange
from rotary_embedding_torch import RotaryEmbedding

class MultiheadAttention(nn.Module):
    r"""
    Args:
        embed_dim: Total dimension of the model.
        num_heads: Number of parallel attention heads. Note that ``embed_dim`` will be split
            across ``num_heads`` (i.e. each head will have dimension ``embed_dim // num_heads``).
        dropout: Dropout probability on ``attn_output_weights``. Default: ``0.0`` (no dropout).
        bias: If specified, adds bias to input / output projection layers. Default: ``True``.
        kdim: Total number of features for keys. Default: ``None`` (uses ``kdim=embed_dim``).
        vdim: Total number of features for values. Default: ``None`` (uses ``vdim=embed_dim``).
    """

    def __init__(self, embed_dim, num_heads, dropout=0., bias=True, 
                 kdim=None, vdim=None, device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        self.scaling = self.head_dim**-0.5

        assert self.head_dim * num_heads == self.embed_dim, "embed_dim must be divisible by num_heads"


        self.q_proj = nn.Linear(embed_dim, embed_dim, bias=False, **factory_kwargs)
        self.k_proj = nn.Linear(embed_dim, self.kdim, bias=False, **factory_kwargs)
        self.v_proj = nn.Linear(embed_dim, self.vdim, bias=False, **factory_kwargs)


        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias, **factory_kwargs)

        self.rotary_emb = RotaryEmbedding(dim=int(self.head_dim/2))

        self._reset_parameters()

    def _reset_parameters(self):
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)

    def forward(
            self,
            query: Tensor,
            key: Tensor,
            value: Tensor,
            attn_mask: Optional[Tensor] = None,
            ) -> Tuple[Tensor, Optional[Tensor]]:

        q, k, v = self.q_proj(query),  self.k_proj(key), self.v_proj(value)

        q = rearrange(q, "b l (h d) -> b h l d", h=self.num_heads, d=self.head_dim)
        k = rearrange(k, "b l (h d) -> b h l d", h=self.num_heads, d=self.head_dim)
        v = rearrange(v, "b l (h d) -> b h l d", h=self.num_heads, d=self.head_dim)

        q = self.rotary_emb.rotate_queries_or_keys(q)
        k = self.rotary_emb.rotate_queries_or_keys(k)
        
        attn_weight = torch.einsum("...ij,...kj->...ik", q, k)/self.scaling

        if attn_mask is not None:
            attn_weight = attn_weight.masked_fill(attn_mask==0, -1e9)
        attn_weight = F.dropout(torch.softmax(attn_weight, dim=-2), p=self.dropout)

        attn = torch.einsum("...ij,...jd->...id", attn_weight, v)

        attn = rearrange(attn, "b h l d -> b l (h d)")
        attn = self.out_proj(attn)

        return attn, attn_weight

class RobertaLMHead(nn.Module):
    """Head for masked language modeling."""

    def __init__(self, embed_dim, output_dim, weight):
        super().__init__()
        self.dense = nn.Linear(embed_dim, embed_dim)
        self.layer_norm = nn.LayerNorm(embed_dim)
        self.weight = weight
        self.bias = nn.Parameter(torch.zeros(output_dim))

    def forward(self, features):
        x = self.dense(features)
        x = F.gelu(x)
        x = self.layer_norm(x)
        # project back to size of vocabulary with bias
        x = F.linear(x, self.weight) + self.bias
        return x
    

class FeedForwardNetwork(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        ffn_embedding_dim: int,
        activation_dropout: float = 0.1,
    ):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.ffn_embedding_dim = ffn_embedding_dim
        self.activation_fn = nn.GELU()
        self.activation_dropout_module = nn.Dropout(
            activation_dropout,
        )
        self.fc1 = nn.Linear(embedding_dim, ffn_embedding_dim)
        self.fc2 = nn.Linear(ffn_embedding_dim, embedding_dim)

    def forward(self, x):
        x = self.activation_fn(self.fc1(x))
        x = self.activation_dropout_module(x)
        x = self.fc2(x)
        return x
    

class TransformerLayer(nn.Module):
    """Transformer layer block."""

    def __init__(
        self,
        embed_dim,
        ffn_embed_dim,
        attention_heads,
        att_dropout=0.,
        ffn_dropout=0.,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.ffn_embed_dim = ffn_embed_dim
        self.attention_heads = attention_heads
        self.att_dropout = att_dropout

        self.mha_layer = MultiheadAttention(
            self.embed_dim,
            self.attention_heads,
            dropout=self.att_dropout,
        )
        self.self_attn_layer_norm = nn.LayerNorm(self.embed_dim)

        self.ffn = FeedForwardNetwork(embed_dim, ffn_embed_dim, ffn_dropout)

        self.final_layer_norm = nn.LayerNorm(self.embed_dim)

    def forward(self, x, self_attn_mask=None):
        residual = x
        x = self.self_attn_layer_norm(x)
        x, attn_weight = self.mha_layer(
            query=x,
            key=x,
            value=x,
            attn_mask=self_attn_mask,
        )
        x = residual + x

        residual = x
        x = self.final_layer_norm(x)
        x = self.ffn(x)
        x = residual + x

        return x, attn_weight

        